import sys

from scenery import logger, config
from scenery.common import Framework
from scenery.cli import parse_args, command
import scenery.commands

def main() -> bool:

    # out: dict[str, dict[str, int | str | dict[str, typing.Any]]] = {}
    success = True
    args = parse_args()

    logger.debug(args)

    # SETUP

    
    if args.command in ["integration", "load"] :
        success &= command(scenery.commands.scenery_setup)(args)
        if config.framework == Framework.DJANGO.value:
            success &= command(scenery.commands.django_setup)(args)

    # MAIN COMMAND

    if args.command == "integration":
        success &= command(scenery.commands.integration_tests)(args)
    elif args.command == "load":
        success &= command(scenery.commands.load_tests)(args)
    elif args.command == "inspect":
        success &= command(scenery.commands.inspect_code)(args)

    if success:
        sys.exit(0)
    else:
        sys.exit(1)


import argparse
import os
import pathlib
import sys
import typing
import dotenv

from scenery import logger, config
from scenery.common import iter_on_manifests
from scenery.cli import (
    interpret, 
    summarize_test_result, 
    report_integration,
    report_load,
    report_inspect
)


########################
# SCENERY CONFIG
########################

def scenery_setup(args: argparse.Namespace) -> bool:
    """Read the settings module and set the corresponding environment variables.
    """

    config.read(args.config)
    dotenv.load_dotenv(args.env)

    logger.debug(dict(config))

    if "manifests" not in config.sections():
        raise ValueError(f"{args.config} should contain a [manifests] section with at least a 'folder' key.")
    
    if "urls" not in config.sections():
        raise ValueError
    
    if args.mode in ["local", "staging", "prod"] and args.mode not in config.urls:
        raise ValueError

    emojy, msg, color, log_lvl = interpret(True)
    logger.info("scenery set-up", style=color)
    
    return True

###################
# DJANGO CONFIG
###################

def django_setup(args: argparse.Namespace) -> bool:
    """Set up the Django environment.

    This function sets the DJANGO_SETTINGS_MODULE environment variable and calls django.setup().

    Args:
        settings_module (str): The import path to the Django settings module.
    """


    import django
    
    sys.path.append(os.path.join('.'))
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", args.django_settings_module)
    logger.debug(f"{os.environ.get('DJANGO_SETTINGS_MODULE')=}")

    django.setup()
    
    from django.conf import settings as django_settings
    logger.debug(f"{django_settings.INSTALLED_APPS=}")

    success = django_settings.configured

    emojy, msg, color, log_lvl = interpret(success)
    logger.log(log_lvl, "django set-up", style=color)
    
    return success


###################
# INTEGRATION TESTS
###################

def integration_tests(args: argparse.Namespace) -> bool:
    """
    Execute the main functionality of the scenery test runner.

    Returns:
        exit_code (int): Exit code indicating success (0) or failure (1)
    """
    # NOTE mad: this needs to be loaded afeter scenery_setup and django_setup
    from scenery.core import process_manifest_as_integration_test


    # FIXME mad: this is here to be able to load driver in two places
    # See also core.TestsLoader.tests_from_manifest.
    # Probably not a great pattern but let's fix this later
    # driver = get_selenium_driver(headless=args.headless)
    driver = None

    report_data  : dict[str, typing.List[typing.Tuple[bool, dict]]] = {
        "dev_backend": [],
        "dev_frontend": [],
        "remote_backend": [],
        "remote_frontend": []
    }


    for filename in iter_on_manifests(args):
        
        results = process_manifest_as_integration_test(filename, args=args, driver=driver)

        for key, val in results.items():
            if val:
                success, summary = summarize_test_result(val, key.replace("_", "-"))
                report_data[key].append((success, summary))

    success = report_integration(report_data)

    return success



###################
# LOAD TESTS
###################

def load_tests(args: argparse.Namespace) -> bool:
    # NOTE mad: this needs to be loaded after scenery_setup and django_setup
    from scenery.core import process_manifest_as_load_test

    success = True
    report_data : dict[str, typing.List] = {}

    for filename in iter_on_manifests(args):        
        results = process_manifest_as_load_test(filename, args=args)
        file_level_success = report_load(results)
        report_data.update(results)
        success &= file_level_success

    return success


###################
# CODE
###################

def inspect_code(args: argparse.Namespace) -> bool:
    from scenery.inspect_code import count_line_types

    report_data = {}

    # Get all files recursively
    folder = pathlib.Path(args.folder)

    # TODO: just the directory (make option)

    # All files
    for file_path in folder.rglob('*.py'):
        if file_path.is_file():
            report_data[str(file_path)] = count_line_types(file_path)

    success = report_inspect(report_data)

    return success





import ast
import pathlib
import tokenize

# NOTE mad: claude wrote this

def count_line_types(file_path: pathlib.Path) -> dict[str, int]:
    """
    Count the number of code lines, docstring lines, and other lines in a Python file.
    
    Args:
        file_path (str): Path to the Python file to analyze
        
    Returns:
        tuple: (code_lines, docstring_lines, other_lines)
    """
    # Read the file content
    with open(file_path, 'r', encoding='utf-8') as file:
        content = file.read()
        
    # Parse the file into an AST
    tree = ast.parse(content)
    
    # Track line numbers for docstrings
    docstring_lines: set[int] = set()
    
    # Extract all docstrings
    extract_docstrings(tree, docstring_lines)
    
    # Tokenize the file to count comments and total lines
    with open(file_path, 'rb') as file:
        tokens = list(tokenize.tokenize(file.readline))
    
    # Get comment lines
    comment_lines = set()
    for token in tokens:
        if token.type == tokenize.COMMENT:
            comment_lines.add(token.start[0])
    
    # Count total lines
    total_lines = max(token.end[0] for token in tokens) if tokens else 0
    
    # Count blank lines
    with open(file_path, 'r', encoding='utf-8') as file:
        blank_lines = set(i+1 for i, line in enumerate(file) 
                            if line.strip() == '')
    
    # Calculate other lines (comments + blank lines)
    other_lines = len(comment_lines.union(blank_lines))
    
    # Calculate code lines (total - docstrings - other)
    code_lines = total_lines - len(docstring_lines) - other_lines
    
    return {"code": code_lines, "docstring": len(docstring_lines), "other": other_lines}


# Helper function to extract docstrings from AST nodes
def extract_docstrings(node: ast.AST, docstring_lines: set) -> set:


    # Check for module, class, and function docstrings
    if isinstance(node, (ast.Module, ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef)):
        docstring = ast.get_docstring(node)
        if docstring:
            # Find line range for this docstring
            if hasattr(node, 'body') and node.body and isinstance(node.body[0], ast.Expr):
                doc_node = node.body[0].value
                if isinstance(doc_node, ast.Constant) and isinstance(doc_node.value, str):
                    start_line = doc_node.lineno
                    end_line = start_line + docstring.count('\n')
                    for line_num in range(start_line, end_line + 1):
                        docstring_lines.add(line_num)
    
    # Recursively process all child nodes
    for child in ast.iter_child_nodes(node):
        extract_docstrings(child, docstring_lines)

    return docstring_lines





"""Build the tests from the Manifest, discover & run tests."""

import argparse
import os
import sys
# import io
from typing import Tuple
import unittest

from scenery import config, logger
from scenery.manifest import Manifest
from scenery.method_builder import MethodBuilder
from scenery.manifest_parser import ManifestParser
from scenery.common import (
    RemoteBackendTestCase,
    RemoteFrontendTestCase,
    DjangoFrontendTestCase,
    DjangoBackendTestCase,
    LoadTestCase,
    Framework,
    # SceneryTestCase,
    get_selenium_driver,
)

from selenium import webdriver

if config.framework == Framework.DJANGO.value:
    from django.conf import settings
    from django.test.utils import get_runner

    from scenery.django_utils import (
        CustomDiscoverRunner,
        DjangoBackendTestCase,
        DjangoFrontendTestCase
    )
    
else:
    class CustomDiscoverRunner:
        def __init__(self, *args, **kwargs):
            raise ValueError(f"Framework {config.framework} is not supported")

# METACLASS
#######################



class MetaTest(type):
    """
    A metaclass for creating test classes dynamically based on a Manifest.

    This metaclass creates test methods for each combination of case and scene in the manifest,
    and adds setup methods to the test class.
    """

    def __new__(
        cls,
        clsname: str,
        bases: tuple[type],
        manifest: Manifest,
        mode: str | None = None,
        only_case_id: str | None = None,
        only_scene_pos: str | None = None,
        only_url: str | None = None,
        driver: webdriver.Chrome | None=None,
        users: int | None=None,
        requests_per_user: int | None=None,
        headless: bool | None=False
    ) -> "MetaTest":
        """Responsible for building the TestCase class.

        Args:
            clsname (str): The name of the class being created.
            bases (tuple): The base classes of the class being created.
            manifest (Manifest): The manifest containing test cases and scenes.

        Returns:
            type: A new test class with dynamically created test methods.

        Raises:
            ValueError: If the restrict argument is not in the correct format.
        """

        # Build setUp and tearDown functions
        ####################################

        # NOTE mad: setUpTestData would be a pain since LoadTestCase and RemoteTestCase 
        # are not suppoesed to work with
        setUpClass = MethodBuilder.build_setUpClass(manifest.set_up_class, driver, headless=headless)
        setUp = MethodBuilder.build_setUp(manifest.set_up)

        cls_attrs = {
            "setUpClass": setUpClass,
            "setUp": setUp,
            "mode": mode,
        }
        if users:
            cls_attrs["users"] = users
        if requests_per_user:
            cls_attrs["requests_per_user"] = requests_per_user

        if  bases in [
            (RemoteFrontendTestCase,),
            (DjangoFrontendTestCase,)
        ]:
            # NOTE mad: used to close the driver
            tearDownClass = MethodBuilder.build_tearDownClass()
            cls_attrs["tearDownClass"] = tearDownClass

        
        # Add test_* functions
        ####################################

        for case_id, scene_pos, take in manifest.iter_on_takes(
            only_url,
            only_case_id,
            only_scene_pos,
        ):
            if bases in [
                (RemoteBackendTestCase,),
                (RemoteFrontendTestCase,),
                (DjangoBackendTestCase,),
                (DjangoFrontendTestCase,),
            ]:
                test = MethodBuilder.build_test_integration(take)
            elif bases == (LoadTestCase,):
                test = MethodBuilder.build_test_load(take)
            else:
                print(bases, bases==(DjangoBackendTestCase,), (DjangoBackendTestCase,))
                raise NotImplementedError(bases)
            cls_attrs.update({f"test_case_{case_id}_scene_{scene_pos}": test})

        test_cls = super().__new__(cls, clsname, bases, cls_attrs)
        return test_cls  


# TEST LOADER AND RUNNER
########################



class TestsRunner:
    """
    A class for running discovered tests and collecting results.

    This class takes discovered tests, runs them using a Django test runner,
    and collects and formats the results.

    Attributes:
        runner (DiscoverRunner): A Django test runner instance.
        logger (Logger): A logger instance for this class.
        discoverer (MetaTestDiscoverer): An instance of MetaTestDiscoverer for discovering tests.
        stream (StringIO): A string buffer for capturing test output.
    """
    # NOTE mad: this was done to potentially shut down the original stream (see also the CustomDiscoverRunner)
    # TODO mad: clarify

    def __init__(self, failfast: bool = False) -> None:
        """Initialize the MetaTestRunner with a runner, logger, discoverer, and output stream."""

        # self.stream = io.StringIO()
        self.stream = sys.stdout
        if config.framework == Framework.DJANGO.value:
            self.runner = CustomDiscoverRunner(stream=self.stream, failfast=failfast)
        else:
            self.runner = unittest.TextTestRunner(stream=self.stream, failfast=failfast)

    def __del__(self) -> None:
        """Clean up resources when the MetaTestRunner is deleted."""
        # NOTE mad: do not erase
        # TODO
        # self.stream.close()
        # print(self.stream.flush())
        # app_logger = ...
        # app_logger.propagate = True

    def run(self, tests_discovered: unittest.TestSuite | unittest.TestCase) -> unittest.TestResult:
        """
        Run the discovered tests and collect results.

        Args:
            tests_discovered (list): A list of tuples, each containing a test name and a TestSuite.
            verbosity (int): The verbosity level for output.

        Returns:
            dict: A dictionary mapping test names to their serialized results.

        Note:
            This method logs test results and prints them to the console based on the verbosity level.
        """
        if isinstance(tests_discovered, unittest.TestCase):
            # NOTE mad: to enforce type, in particular for type checking
            tests_discovered = unittest.TestSuite((tests_discovered,))
        if isinstance(self.runner, CustomDiscoverRunner):
            # NOTE mad: I stick to the method name used by the original django class
            # this sounds more explicit than a addition of a run method somewhere else
            results = self.runner.run_suite(tests_discovered)
        else:
            results = self.runner.run(tests_discovered)
        return results




class TestsDiscoverer:
    """
    A class for discovering and loading test cases from manifest files.

    This class scans a directory for manifest files, creates test classes from these manifests,
    and loads the tests into test suites.

    Attributes:
        logger (Logger): A logger instance for this class.
        runner (DiscoverRunner): A Django test runner instance.
        loader (TestLoader): A test loader instance from the runner.
    """

    if config.framework == Framework.DJANGO.value:
        runner = get_runner(settings, test_runner_class="django.test.runner.DiscoverRunner")()
        loader: unittest.loader.TestLoader = runner.test_loader

    else:
        runner = unittest.TextTestRunner()
        loader = unittest.loader.TestLoader()


    def integration_tests_from_manifest(
        self,
        filename: str,
        mode: str,
        back: bool = False,
        front: bool = False,
        only_url: str | None = None,
        # timeout_waiting_time: int = 5,
        only_case_id: str | None = None,
        only_scene_pos: str | None = None,
        driver: webdriver.Chrome | None = None,
        headless: bool = False,
        page_load_timeout: int= 30,
    ) -> Tuple[unittest.TestSuite, unittest.TestSuite, unittest.TestSuite, unittest.TestSuite]:
        """Creates test suites from a manifest file for both backend and frontend testing.

        Parses a YAML manifest file and generates corresponding test suites for backend
        and frontend testing. Tests can be filtered based on various criteria like test type,
        specific views, or test cases.

        Args:
            filename (str): The name of the YAML manifest file to parse.
            only_back (bool, optional): Run only backend tests. Defaults to False.
            only_front (bool, optional): Run only frontend tests. Defaults to False.
            only_url (str, optional): Filter tests to run only for a specific view. Defaults to None.
            timeout_waiting_time (int, optional): Timeout duration for frontend tests in seconds. Defaults to 5.
            only_case_id (str, optional): Filter tests to run only for a specific case ID. Defaults to None.
            only_scene_pos (str, optional): Filter tests to run only for a specific scene position. Defaults to None.
            driver (webdriver.Chrome, optional): Selenium Chrome WebDriver instance. If None, creates new instance. Defaults to None.
            headless (bool, optional): Whether to run browser in headless mode. Defaults to True.

        Returns:
            Tuple[unittest.TestSuite, unittest.TestSuite]: A tuple containing:
                - Backend test suite (first element)
                - Frontend test suite (second element)

        Notes:
            - The manifest's ttype determines which suites are created (backend, frontend, or both)
            - Empty test suites are returned for disabled test types
            - The driver initialization can occur here or be passed in from external code
        """

        dev_backend_suite, dev_frontend_suite = unittest.TestSuite(), unittest.TestSuite()
        remote_backend_suite, remote_frontend_suite = unittest.TestSuite(), unittest.TestSuite()

        # Parse manifest
        manifest = ManifestParser.parse_yaml_from_file(os.path.join(config.manifests_folder, filename))
        ttype = manifest.ttype
        manifest_name = filename.replace(".yml", "")

        # NOTE mad: manifests can indicate they are not to be ran in some mode
        back &= ttype is None or ttype == "backend"
        front &= ttype is None or ttype == "frontend"

        if back and mode == "dev":
            cls = MetaTest(
                f"{manifest_name}.dev.backend",
                (DjangoBackendTestCase,),
                manifest,
                only_case_id=only_case_id,
                only_scene_pos=only_scene_pos,
                only_url=only_url,
                mode=mode,
            )
            # FIXME mad: type hinting mislead by metaclasses
            tests = self.loader.loadTestsFromTestCase(cls)  # type: ignore[arg-type]
            dev_backend_suite.addTests(tests)

        if back and mode in ["local", "staging", "prod"]:
            cls = MetaTest(
                f"{manifest_name}.remote.backend",
                (RemoteBackendTestCase,),
                manifest,
                only_case_id=only_case_id,
                only_scene_pos=only_scene_pos,
                only_url=only_url,
                mode=mode
            )

            # FIXME mad: type hinting mislead by metaclasses
            tests = self.loader.loadTestsFromTestCase(cls) # type: ignore[arg-type]
            remote_backend_suite.addTests(tests)

        # Create frontend test
        if front and mode == "dev":
            # NOTE mad: this is here to be able to load driver in two places
            # See also scenery/__main__.py
            # Probably not a great pattern but let's FIXME this later
            if driver is None:
                driver = get_selenium_driver(headless=headless, page_load_timeout=page_load_timeout)

            cls = MetaTest(
                f"{manifest_name}.dev.frontend",
                (DjangoFrontendTestCase,),
                manifest,
                only_case_id=only_case_id,
                only_scene_pos=only_scene_pos,
                only_url=only_url,
                driver=driver,
                mode=mode,
                headless=headless,
            )
            # FIXME mad: type hinting mislead by metaclasses
            tests = self.loader.loadTestsFromTestCase(cls)  # type: ignore[arg-type]
            dev_frontend_suite.addTests(tests)


        if front and mode in ["local", "staging", "prod"]:
            cls = MetaTest(
                f"{manifest_name}.remote.frontend",
                (RemoteFrontendTestCase,),
                manifest,
                only_case_id=only_case_id,
                only_scene_pos=only_scene_pos,
                only_url=only_url,
                driver=driver,
                mode=mode,
                headless=headless,
            )

            # FIXME mad: type hinting mislead by metaclasses
            tests = self.loader.loadTestsFromTestCase(cls) # type: ignore[arg-type]
            remote_frontend_suite.addTests(tests)

        return dev_backend_suite, dev_frontend_suite, remote_backend_suite, remote_frontend_suite


    def load_tests_from_manifest(
        self,
        filename: str,
        mode: str,
        users: int,
        requests_per_user: int,
        only_url: str | None = None,
        only_case_id: str | None = None,
        only_scene_pos: str | None = None,
    ) -> unittest.TestSuite:

        test_suite = unittest.TestSuite()

        # Parse manifest
        manifest = ManifestParser.parse_yaml_from_file(os.path.join(config.manifests_folder, filename))
        manifest_name = filename.replace(".yml", "")

        cls = MetaTest(
            f"{manifest_name}.load",
            (LoadTestCase,),
            manifest,
            only_case_id=only_case_id,
            only_scene_pos=only_scene_pos,
            only_url=only_url,
            mode=mode,
            users=users,
            requests_per_user=requests_per_user
        )

        # FIXME mad: type hinting mislead by metaclasses
        tests = self.loader.loadTestsFromTestCase(cls)  # type: ignore[arg-type]
        test_suite.addTests(tests)

        return test_suite



def process_manifest_as_integration_test(
    manifest_filename: str, args: argparse.Namespace, driver: webdriver.Chrome | None
) -> dict:
    """Process a test manifest file and executes both backend and frontend tests.

    Takes a manifest file and command line arguments to run the specified tests,
    collecting and summarizing the results for both backend and frontend test suites.

    Args:
        filename (str): The name of the YAML manifest file to process.
        args (argparse.Namespace): Command line arguments containing:
            - only_back (bool): Run only backend tests
            - only_front (bool): Run only frontend tests
            - only_url (str): Filter for specific view
            - only_case_id (str): Filter for specific case ID
            - only_scene_pos (str): Filter for specific scene position
            - timeout_waiting_time (int): Frontend test timeout duration
            - headless (bool): Whether to run browser in headless mode
        driver (webdriver.Chrome | None): Selenium Chrome WebDriver instance or None.

    Returns:
        Tuple[bool, dict, bool, dict]: A tuple containing:
            - Backend test success status (bool)
            - Backend test summary results (dict)
            - Frontend test success status (bool)
            - Frontend test summary results (dict)

    Notes:
        - Prints the manifest name (without .yml extension) during execution
        - Uses TestsLoader and TestsRunner for test execution
        - Test results are summarized with verbosity level 0
    """

    # logging.log(logging.INFO, f"{manifest_filename=}")
    logger.info(f"{manifest_filename=}")

    loader = TestsDiscoverer()
    runner = TestsRunner()

    dev_backend_suite, dev_frontend_suite, remote_backend_suite, remote_frontend_suite = (
        loader.integration_tests_from_manifest(
            manifest_filename,
            mode=args.mode,
            back=args.back,
            front=args.front,
            only_url=args.url,
            only_case_id=args.case_id,
            only_scene_pos=args.scene_pos,
            # timeout_waiting_time=args.timeout_waiting_time,
            driver=driver,
            headless=args.headless,
            page_load_timeout=args.page_load_timeout,
        )
    )

    results: dict[str, unittest.TestResult | None] = {
        "dev_backend": None,
        "dev_frontend": None,
        "remote_backend": None,
        "remote_frontend": None,
    }

    if args.back and args.mode == "dev":
        results["dev_backend"] = runner.run(dev_backend_suite)

    if args.back and args.mode in ["local", "staging", "prod"]:
        results["remote_backend"] = runner.run(remote_backend_suite)

    if args.front and args.mode == "dev":
        results["dev_frontend"] = runner.run(dev_frontend_suite)

    if args.front and args.mode in ["local", "staging", "prod"]:
        results["remote_frontend"] = runner.run(remote_frontend_suite)

    return results


def process_manifest_as_load_test(
    manifest_filename: str, args: argparse.Namespace
) -> dict:
    results = {}

    logger.info(f"{manifest_filename=}")

    loader = TestsDiscoverer()
    runner = TestsRunner()

    tests_suite = loader.load_tests_from_manifest(
        manifest_filename,
        mode=args.mode,
        only_url=args.url,
        only_case_id=args.case_id,
        only_scene_pos=args.scene_pos,
        users=args.users,
        requests_per_user=args.requests,
    )

    for test in tests_suite:
        # NOTE mad: we build the suite in order to satisfy this
        assert isinstance(test, LoadTestCase)

        test_result = runner.run(test)
        if len(test_result.errors) == 0:
            results.update(test.data)
        else:
            # print(test_result.errors)
            raise Exception

    return results






"""Execute instructions used in `TestCase.setUp` and `TestCase.setUpTestData` provided in the manifest."""

import importlib
from typing import Callable

from scenery import config, logger
from scenery.common import SceneryTestCase, RemoteBackendTestCase, LoadTestCase, RemoteFrontendTestCase
from scenery.manifest import SetUpInstruction


def local_execution_only(func: Callable) -> Callable:
    """Simple decorator that marks function as executable"""
    # NOTE mad: this "ignore" is here to stay, the whole
    # point is to create a non existing attribute.
    func._local_execution_only = True # type: ignore [attr-defined]
    return func

class SetUpHandler:
    """Responsible for executing instructions used in `TestCase.setUp` and `TestCase.setUpTestData` provided in the manifest.

    This class dynamically imports and executes setup instructions specified in the test manifest.
    It is typically used by the MethodBuilder to construct setup methods for test cases.

    Attributes:
        module: The imported module containing setup instruction implementations.
        logger: A logger instance for debug output.
    """


    set_up_instructions = None

    @classmethod
    def load_set_up_instructions(cls):
        instructions_file_path = config.setup_instructions
        if instructions_file_path is not None and cls.set_up_instructions is None:
            spec = importlib.util.spec_from_file_location("dynamic_module", instructions_file_path)
            module = importlib.util.module_from_spec(spec)
            # sys.modules[module_name] = module  # Optional: add to sys.modules
            spec.loader.exec_module(module)
            cls.set_up_instructions = module

    
    @classmethod
    def exec_set_up_instruction(
        cls,
        # NOTE mad: it either takes the instance or the class
        # depending whether it is class method or not
        # (setUp vs. setUpClass)
        testcase: SceneryTestCase | type[SceneryTestCase],
        instruction: SetUpInstruction,
    ) -> None:
        """Execute the method corresponding to the SetUpInstruction.

        This method dynamically retrieves and executes the setup function specified
        by the SetUpInstruction. It logs the execution for debugging purposes.

        Args:
            testcase (SceneryTestCase): The Scenery test case class or instance.
            instruction (scenery.manifest.SetUpInstruction): The setup instruction to execute.

        Raises:
            AttributeError: If the specified setup function is not found in the imported module.
        """

        if cls.set_up_instructions is None:
            cls.load_set_up_instructions()

        func = getattr(cls.set_up_instructions, instruction.command)

        if isinstance(testcase, (RemoteBackendTestCase, LoadTestCase, RemoteFrontendTestCase)) and hasattr(func, "_local_execution_only"):
            pass
        else:
            logger.debug(instruction)
            func(testcase, **instruction.args)




"""Building test methods dynamically based on manifest data."""
import collections
import requests
from typing import Callable
import threading
import time
import http

from scenery import config, logger
from scenery.manifest import SetUpInstruction, Take, DirectiveCommand
from scenery.response_checker import Checker, ResponseProtocol
from scenery.set_up_handler import SetUpHandler
from scenery.common import (
    SceneryTestCase,
    RemoteBackendTestCase,
    RemoteFrontendTestCase,
    DjangoBackendTestCase,
    DjangoFrontendTestCase,
    LoadTestCase,
    get_selenium_driver,
    Framework
)

from selenium import webdriver

if config.framework == Framework.DJANGO.value:
    from scenery.django_utils  import DjangoBackendTestCase, DjangoFrontendTestCase

################
# METHOD BUILDER
################

class MethodBuilder:
    """A utility class for building test methods dynamically based on manifest data.

    This class provides static methods to create setup and test methods
    that can be added to Django test cases.
    """

    # NOTE mad: do not erase, but this is unused right now, but tested
    @staticmethod
    def build_setUpTestData(instructions: list[SetUpInstruction]) -> classmethod:
        """Build a setUpTestData class method for a Django test case.

        This method creates a class method that executes a series of setup
        instructions before any test methods are run.

        Args:
            instructions (list[str]): A list of setup instructions to be executed.

        Returns:
            classmethod: A class method that can be added to a Django test case.
        """

        def setUpTestData(testcase_cls: type[SceneryTestCase]) -> None:
            # TODO mad: not sure why this is needed
            super(testcase_cls, testcase_cls).setUpTestData() # type: ignore[misc]

            for instruction in instructions:
                SetUpHandler.exec_set_up_instruction(testcase_cls, instruction)

        return classmethod(setUpTestData)

    @staticmethod
    def build_setUpClass(
        instructions: list[SetUpInstruction], driver: webdriver.Chrome | None, headless: bool = True
    ) -> classmethod:
        """
        Build and return a class method for setup operations before any tests in a test case are run.

        This method generates a setUpClass that:
        - Sets up the test environment using Django's setup
        - For FrontendDjangoTestCase subclasses, initializes a Selenium WebDriver
        - Executes a list of setup instructions

        Args:
            instructions: A list of SetUpInstruction objects to be executed during setup
            driver: An optional pre-configured Selenium Chrome WebDriver. If None, a new driver
                will be created with the specified headless setting
            headless: Boolean flag to run Chrome in headless mode (default: True)

        Returns:
            classmethod: A class method that handles test case setup operations
        """

        def setUpClass(testcase_cls: type[SceneryTestCase]) -> None:
            logger.debug(setUpClass.__name__)
            # TODO mad: not sure why this is needed
            super(testcase_cls, testcase_cls).setUpClass()  # type: ignore[misc]

            if issubclass(testcase_cls, (RemoteFrontendTestCase, DjangoFrontendTestCase,)):
                if driver is None:
                    testcase_cls.driver = get_selenium_driver(headless)
                else:
                    testcase_cls.driver = driver

                # chrome_options = Options()
                # # NOTE mad: service does not play well with headless mode
                # # service = Service(executable_path='/usr/bin/google-chrome')
                # if headless:
                #     chrome_options.add_argument("--headless=new")     # NOTE mad: For newer Chrome versions
                #     # chrome_options.add_argument("--headless")           # NOTE mad: For older Chrome versions (Framework)
                # testcase_cls.driver = webdriver.Chrome(options=chrome_options) #  service=service
                # testcase_cls.driver.implicitly_wait(10)

            for instruction in instructions:
                SetUpHandler.exec_set_up_instruction(testcase_cls, instruction)

        return classmethod(setUpClass)

    @staticmethod
    def build_tearDownClass() -> classmethod:
        """
        Build and return a class method for teardown operations after all tests in a test case have completed.

        The generated tearDownClass method performs cleanup operations, specifically:
        - For FrontendDjangoTestCase subclasses, it quits the Selenium WebDriver
        - Calls the parent class's tearDownClass method

        Returns:
            classmethod: A class method that handles test case teardown operations
        """

        def tearDownClass(testcase_cls: type[SceneryTestCase]) -> None:
            if issubclass(testcase_cls, DjangoFrontendTestCase):
                testcase_cls.driver.quit()
            # TODO mad: not sure why this is needed

            super(testcase_cls, testcase_cls).tearDownClass()  # type: ignore[misc]

        return classmethod(tearDownClass)
    

    @staticmethod
    def build_setUp(
        instructions: list[SetUpInstruction]
    ) -> Callable[[SceneryTestCase], None]:
        """Build a setUp instance method for a Django test case.

        This method creates an instance method that executes a series of setup
        instructions before each test method is run.

        Args:
            instructions (list[str]): A list of setup instructions to be executed.

        Returns:
            function: An instance method that can be added to a Django test case.
        """
        def setUp(testcase: SceneryTestCase) -> None:

            logger.debug(setUp.__name__)

            # if isinstance(testcase, (RemoteBackendTestCase, LoadTestCase, DjangoFrontendTestCase)):
            #     testcase.session = requests.Session()
            if isinstance(testcase, (RemoteBackendTestCase, LoadTestCase,)):
                testcase.session = requests.Session()
                testcase.headers = {}
            if isinstance(testcase, (RemoteBackendTestCase, RemoteFrontendTestCase, LoadTestCase)) :
                testcase.base_url = config["urls"][testcase.mode]
            if isinstance(testcase, (DjangoFrontendTestCase,)) :
                testcase.base_url = testcase.live_server_url
            if isinstance(testcase, (LoadTestCase,)):
                testcase.data = collections.defaultdict(list)


            for instruction in instructions:
                SetUpHandler.exec_set_up_instruction(testcase, instruction)

        return setUp



    @staticmethod
    def build_test_integration(take: Take) -> Callable:
        """Build a test method from an Take object.

        This method creates a test function that sends an HTTP request
        based on the take's specifications and executes a series of checks
        on the response.

        Args:
            take (scenery.manifest.Take): An Take object specifying
                the request to be made and the checks to be performed.

        Returns:
            function: A test method that can be added to a Django test case.
        """

        def test(testcase: SceneryTestCase) -> None:
            logger.debug(test.__name__)

            response: ResponseProtocol
            

            if isinstance(testcase, RemoteBackendTestCase):
                response = Checker.get_http_response(testcase, take)

            elif isinstance(testcase, DjangoBackendTestCase):
                response = Checker.get_django_client_response(testcase, take)

            elif isinstance(testcase,(RemoteFrontendTestCase, DjangoFrontendTestCase) ):
                response = Checker.get_selenium_response(testcase, take)

            else:
                raise ValueError(f"Unsupported test case type: {type(testcase)}")
            

            for i, check in enumerate(take.checks):

                if isinstance(testcase, RemoteBackendTestCase) and check.instruction in [
                        DirectiveCommand.COUNT_INSTANCES,
                        DirectiveCommand.FIELD_OF_INSTANCE,
                    ]:
                        continue
                if isinstance(testcase, (RemoteFrontendTestCase, DjangoFrontendTestCase)) and check.instruction == DirectiveCommand.STATUS_CODE:
                    continue


                with testcase.subTest(f"directive {i}"):
                    Checker.exec_check(testcase, response, check)

        return test

    @staticmethod
    def build_test_load(take: Take) -> Callable:
        def test(testcase: LoadTestCase) -> None:
            
            lock = threading.Lock()  # Thread synchronization

            def make_request(testcase: LoadTestCase, session: requests.Session, take: Take, headers: dict[str, str]) -> dict[str, int|float]:
                """Execute a single request and return response time and status"""

                start_time = time.time()

                if take.method == http.HTTPMethod.GET:
                    response = session.get(
                        testcase.base_url + take.url,
                        data=take.data,
                        headers=headers,
                    )
                elif take.method == http.HTTPMethod.POST:
                    response = session.post(
                        testcase.base_url + take.url,
                        take.data,
                        headers=headers,
                    )
                else:
                    raise NotImplementedError(take.method)

                        
                elapsed_time = time.time() - start_time

                # print(response.status_code)


                if not (200 <= response.status_code < 300):
                    logger.warning(f"{response.status_code=}")
                    logger.debug(f"{response.content.decode('utf8')=}")
                return {
                    'elapsed_time': elapsed_time,
                    'status_code': response.status_code,
                    'success': 200 <= response.status_code < 300
                }

            def _worker_task(testcase: LoadTestCase, take: Take, num_requests: int) -> None:
                """Worker function executed by each thread"""
                for _ in range(num_requests):
                    result = make_request(testcase, testcase.session, take, testcase.headers)
                    
                    with lock:
                        testcase.data[take.url].append(result)

            # logger.info(f"{ramp_up=}")
            logger.info(f"{testcase.users=}")
            logger.info(f"{testcase.requests_per_user=}")
            logger.info(f"{take.url=}")
            logger.info(f"{take.method=}")
            logger.info(f"{take.data=}")

            # Create threads for each simulated user
            threads = []
            for i in range(testcase.users):
                thread = threading.Thread(
                    target=_worker_task,
                    args=(testcase, take, testcase.requests_per_user),
                )
                threads.append(thread)

                # Optional: implement ramp-up by staggering thread starts
                thread.start()
                # if ramp_up > 0 and users > 1:
                #     time.sleep(ramp_up / (users - 1))

            # Wait for all threads to complete
            for thread in threads:
                thread.join()


        return test


from typing import IO, Any

from django.test.runner import DiscoverRunner as DjangoDiscoverRunner
import django.test
from django.contrib.staticfiles.testing import StaticLiveServerTestCase

from selenium import webdriver

class DjangoBackendTestCase(django.test.TestCase):
    """A Django TestCase for backend testing."""

class DjangoFrontendTestCase(StaticLiveServerTestCase):
    """A Django TestCase for frontend testing."""
    base_url: str
    driver: webdriver.Chrome


def overwrite_get_runner_kwargs(
    django_runner: DjangoDiscoverRunner, stream: IO
) -> dict[str, Any]:
    """Overwrite the get_runner_kwargs method of Django's DiscoverRunner.

    This function is used to avoid printing Django test output by redirecting the stream.

    Args:
        django_runner (DiscoverRunner): The Django test runner instance.
        stream: The stream to redirect output to.

    Returns:
        dict: A dictionary of keyword arguments for the test runner.

    Notes:
        see django.test.runner.DiscoverRunner.get_runner_kwargs
    """
    kwargs = {
        "failfast": django_runner.failfast,
        "resultclass": django_runner.get_resultclass(),
        "verbosity": django_runner.verbosity,
        "buffer": django_runner.buffer,
        # NOTE: this is the line below that changes compared to the original
        "stream": stream,
    }
    return kwargs


class CustomDiscoverRunner(DjangoDiscoverRunner):
    """Custom test runner that allows for stream capture."""
    
    # NOTE mad: this was done to potentially shut down the original stream
    # NOTE mad: used both in rehearsal and core module (for the runner
    # TODO mad: once we fit FastAPI, this runner should only be used 
    # for the django backend and frontend test

    def __init__(self, stream: Any , *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.stream = stream

    # def __del__(self):
    #     print(self.stream.getvalue())


    def get_test_runner_kwargs(self) -> dict[str, Any]:
        """Overwrite the original from django.test.runner.DiscoverRunner."""
        return overwrite_get_runner_kwargs(self, self.stream)





import argparse
import collections
import logging
import statistics
import unittest
from typing import Any, Callable, Iterable, Tuple

from scenery import logger, console
from scenery.common import serialize_unittest_result

from rich import box
from rich.console import Console
from rich.rule import Rule
from rich.progress import Progress, BarColumn, TextColumn
from rich.style import Style
from rich.table import Table
from rich.text import Text
from rich.columns import Columns
from rich.console import Group
from rich.panel import Panel



##################
# COMMAND WRAPPER
##################


def command(func: Callable) -> Callable:
    def wrapper(*args: Any) -> Any:

        command_label = func.__name__.replace("_", " ").capitalize()

        console.print(Rule(f"[section]{command_label} start...[/section]", style="cyan"))
        # logger.info(f"starting {func.__name__}...")

        success = func(*args)

        emojy, msg, color, log_lvl = interpret(success)
        console.print(Rule(f"{emojy} {command_label} {msg}", style=color))

        return success

    return wrapper


#################
# PARSE ARGUMENTS
#################


def parse_args() -> argparse.Namespace:
    """Parse command line arguments with subcommands."""
    parser = argparse.ArgumentParser(description="Scenery Testing Framework")

    # Add subparsers
    subparsers = parser.add_subparsers(dest="command", help="Testing command to run")
    parse_integration_args(subparsers)
    parse_load_args(subparsers) 
    parse_inspect_args(subparsers) 

    args = parser.parse_args()

    if hasattr(args, "test"):
        args.manifest, args.case_id, args.scene_pos = parse_arg_test_restriction(args.test)

    logger.level = logging._nameToLevel[args.log]

    return args


def add_common_testing_arguments(parser: argparse.ArgumentParser) -> None:

    parser.add_argument(
        "--config",
        dest="config",
        type=str,
        default="scenery.ini",
        help="Location of scenery settings module",
    )

    parser.add_argument(
        "--env",
        dest="env",
        type=str,
        default=".env",
        help="Location of env variables (for test credentials)",
    )

    # parser.add_argument(
    #     "--port",
    #     dest="port",
    #     type=str,
    #     default="8020",
    #     help="Port on which the server runs",
    # )

    parser.add_argument(
        "--django-settings",
        dest="django_settings_module",
        type=str,
        default=None,
        help="Location of django settings module, if needed",
    )

    parser.add_argument(
        "--test",
        nargs="?",
        default=None,
        help="Optional test restriction <manifest>.<case>.<scene>",
    )

    parser.add_argument(
        "--url",
        nargs="?",
        default=None,
        help="Optional url restriction",
    )

    parser.add_argument(
        "--log",
        nargs="?",
        default="INFO",
        help="Log level",
    )


def parse_arg_test_restriction(test_name_pattern: str|None) -> Tuple[str|None, str|None, str|None]:
    """Parse the --test argument into a tuple of (manifest_name, case_id, scene_pos)."""
    if test_name_pattern is not None:
        split_pattern = test_name_pattern.split(".")
        if len(split_pattern) == 1:
            manifest_name, case_id, scene_pos = split_pattern[0], None, None
        elif len(split_pattern) == 2:
            manifest_name, case_id, scene_pos = split_pattern[0], split_pattern[1], None
        elif len(split_pattern) == 3:
            manifest_name, case_id, scene_pos = split_pattern[0], split_pattern[1], split_pattern[2]
        else:
            raise ValueError(f"Wrong restrict argmuent {test_name_pattern}")
        return manifest_name, case_id, scene_pos
    else:
        return None, None, None


def parse_integration_args(subparser: argparse._SubParsersAction) -> None:
    """Parse command line arguments."""

    parser = subparser.add_parser('integration', help='Integration tests')
    add_common_testing_arguments(parser)
    
    parser.add_argument('--back', action='store_true')
    parser.add_argument('--front', action='store_true')

    parser.add_argument(
        "--mode",
        choices=["dev", "local", "staging", "prod"],
        default="dev"
    )

    # parser.add_argument(
    #     "--timeout",
    #     dest="timeout_waiting_time",
    #     type=int,
    #     default=5,
    # )

    parser.add_argument(
        "--page-load-timeout",
        dest="page_load_timeout",
        type=int,
        default=5,
    )

    parser.add_argument('--headless', action='store_true')
    parser.add_argument('--failfast', action='store_true')



def parse_load_args(subparser: argparse._SubParsersAction) -> None:

    parser = subparser.add_parser('load', help='Load tests')
    add_common_testing_arguments(parser)


    parser.add_argument(
        "--mode",
        choices=["local", "staging", "prod"],
    )

    parser.add_argument('-u', '--users', type=int)
    parser.add_argument('-r', '--requests', type=int)

def parse_inspect_args(subparser: argparse._SubParsersAction) -> None:

    parser = subparser.add_parser('inspect', help='Inspect files')

    parser.add_argument('-f', '--folder', type=str, help='Folder to inspect')

    parser.add_argument(
        "--log",
        nargs="?",
        default="INFO",
        help="Log level",
    )

#################
# RICH HELPERS
#################


def table_from_dict(d: dict[str, Any], col1_title : str = "", col2_title:str = "", title: str | None =None, formatting: dict[str, Tuple[str, Any]]={}) -> Table:
        
    show_header = col1_title != "" or col2_title != ""
    table = Table(title=title, box=box.ROUNDED, show_header=show_header)
    table.add_column(col1_title, style="cyan", no_wrap=True)
    table.add_column(col2_title, justify="right")

    for key, value in d.items():

        format_str, color = formatting.get(key, ("{}", None))
        label = str(key).replace("_", " ").capitalize()
        row_values = [label,]
        value = d.get(key)
        formatted_value = format_str.format(value)
        if color:
            formatted_value = f"[{color}]{formatted_value}[/{color}]"
        row_values.append(formatted_value)
        
        table.add_row(*row_values)
    
    return table

def histogram(x: Iterable[int]) -> Progress:
    """Display a histogram leveraging Rich progress bars and return a renderable object"""
    # Calculate histogram data
    max_val = max(x)
    
    bins = [
        (0, 50), 
        (50, 100), 
        (100, 200),
        (100, 300),
        (300, 500),
        (500, 750),
        (750, 1000),
        (1000, 2000),
        (2000, 3000),
        (3000, max(max_val, 4000))]

    histogram_data = []
    # for i in range(bins):
        # bin_start = min_val + i * bin_width
        # bin_end = min_val + (i + 1) * bin_width
    for i, (bin_start, bin_end) in enumerate(bins):
        bin_count = sum(1 for t in x if bin_start <= t < bin_end or (i == len(bins)-1 and t == bin_end))
        # bin_count = sum(1 for t in x if bin_start <= t < bin_end or (i == bins-1 and t == bin_end))
        histogram_data.append((bin_start, bin_end, bin_count))
    
    max_count = sum(count for _, _, count in histogram_data) if histogram_data else 0
    completed_style = Style(color="white")
    
    # Create a custom Progress object that can be rendered later
    progress = Progress(
        TextColumn("[bold]{task.description}"),
        BarColumn(bar_width=50, complete_style=completed_style),
        TextColumn("{task.completed} requests"),
        expand=False,
    )
    
    # Add tasks to the progress object
    for bin_start, bin_end, count in histogram_data:
        desc = f"{bin_start:.0f}ms - {bin_end:.0f}ms"
        progress.add_task(desc, total=max_count, completed=count)
    
    # Return the progress object itself (which is renderable)
    return progress




##################
# REPORTS
##################


def summarize_test_result(result: unittest.TestResult, test_label: str) -> Tuple[bool, collections.Counter]:
    """Return true if the tests all succeeded, false otherwise."""
    for failed_test, traceback in result.failures:
        test_name = failed_test.id()
        emojy, msg, color, log_lvl = interpret(False)
        # logger.log(log_lvl, f"[{color}]{test_name} {msg}[/{color}]\n{traceback}")
        logger.log(log_lvl, f"{test_name} {msg}", style=color)
        # console.print_exception(
        #     exc_info=(None, None, traceback),
        #     show_locals=True
        # )
        # console.print(traceback)

    for failed_test, traceback in result.errors:
        test_name = failed_test.id()
        emojy, msg, color, log_lvl = interpret(False)
        # logger.log(log_lvl, f"[{color}]{test_name} {msg}[/{color}]\n{traceback}")
        logger.log(log_lvl, f"{test_name} {msg}", style=color)
        # console.print(traceback)

    success = True
    summary = serialize_unittest_result(result)
    if summary["errors during execution"] > 0 or summary["failed assertions"] > 0:
        success = False
    # if summary["testsRun"] == 0:
    #     pass

    else:
        emojy, msg, color, log_lvl = interpret(success)
        msg = f"{test_label} {msg}"
        logger.log(log_lvl, msg, style=color)

    return success, summary


def interpret(success: bool) -> Tuple[str, str, str, int]:
    """Return emojy, success/failure message, color and log level corresponding to the succes of an execution."""
    if success:
        emojy, msg, color, log_lvl = "🟢", "passed", "green", logging.INFO
    else:
        emojy, msg, color, log_lvl = "❌", "failed", "red", logging.ERROR
    return emojy, msg, color, log_lvl



def report_integration(data: dict) -> bool:


    panel_msg = ""
    panel_color = "green"
    report_tables = []

    command_level_success = True

    for key, val in data.items():

        key_level_success = True
        key_level_summary : dict[str, int] = collections.Counter()

        for success, summary in val:
            
            key_level_success &= success
            key_level_summary.update(summary)

        if val:
            emojy, msg, color, log_lvl = interpret(key_level_success)

            if key_level_success:
                msg = f"all {key} tests {msg}"
            else:
                msg = f"some {key} tests {msg}"
                panel_color = "red"

            logger.log(log_lvl, msg, style=color)
            if panel_msg != "":
                panel_msg += "\n"
            panel_msg += f"{emojy} {msg}"
            report_tables.append(table_from_dict(key_level_summary, key, ""))

            command_level_success &= key_level_success

    emojy, msg, color, log_lvl = interpret(command_level_success)
    logger.log(log_lvl, f"integration tests {msg}", style=color)

    fmt_tables = Columns(report_tables, equal=False, expand=True)
    panel_report = Group(panel_msg, fmt_tables)

    console.print(Panel(panel_report, title="Results", border_style=panel_color))


    return command_level_success



def report_load(data: dict, threshold_p95: int = 500, threshold_p99: int = 5000) -> bool:

    #####################
    # OUTPUT
    #####################

    command_level_success = True


    console = Console()

    for endpoint, requests_results in data.items():

        total_requests = len(requests_results)
        if total_requests == 0:
            continue

        successes = [r for r in requests_results if r["success"]]
        failures = [r for r in requests_results if not r["success"] ]

        success_times = [r['elapsed_time']*1000 for r in successes]
        error_times = [r['elapsed_time']*1000 for r in failures]
        
        error_rate = (len(error_times) / total_requests) * 100 if total_requests > 0 else 0
        
        ep_analysis = {
            'total_requests': total_requests,
            'successful_requests': len(success_times),
            'failed_requests': len(error_times),
            'error_rate': error_rate
        }


        
        if success_times:
            quantiles = statistics.quantiles(success_times, n=100)
            # quantiles = statistics.quantiles(success_times, n=4)

            p50 = statistics.median(success_times)
            p90 = quantiles[90-1]
            p95 = quantiles[95-1]
            p99 = quantiles[99-1]

            ep_analysis.update({
                'min_time': min(success_times),
                'p50': p50,
                'p90': p90,
                'p95': p95,
                '[bold]p99[/bold]': p99,
                'max': max(success_times),

            })
            
            if len(success_times) > 1:
                ep_analysis['stdev'] = statistics.stdev(success_times)

            # TODO mad: confirm with sel
            success = bool(p95 < threshold_p95)
            success &= bool(p99 < threshold_p99)

        else:
            success = False
        
        command_level_success &= success

        formatting = {
            "error_rate": ("{:.2f}%", None),
            "min_time": ("{:.2f}ms", None),
            "p50": ("{:.2f}ms", None),
            "p90": ("{:.2f}ms", None),
            "p95": ("{:.2f}ms", None),
            "[bold]p99[/bold]": ("[bold]{:.2f}ms[/bold]", None),
            "max": ("{:.2f}ms", None),
            "stdev": ("{:.2f}ms", None),
        }
         

        table = table_from_dict(
            ep_analysis, 
            "Metric", 
            "Value", 
            "",
            formatting,
            )
        
        plot = histogram(success_times)

        fmt_plot = Group(Text("\n"*1), plot)
        successes_columns = Columns([table, fmt_plot], equal=False, expand=True)

        if failures:
            failures_status_codes = collections.Counter([r["status_code"] for r in failures])        
            failures_table = table_from_dict(
                failures_status_codes, 
                "Status code of failed requests", 
                "N", 
                "",
                )
            panel_content = Group(successes_columns, failures_table)

        else:
            panel_content = Group(successes_columns)

        console.print(Panel(panel_content, title=f"{endpoint=}"))
    
    return command_level_success



def report_inspect(data: dict, code_threshold: int=300) -> bool:

    show_header = True
    table = Table(title="Line count", box=box.ROUNDED, show_header=show_header)
    table.add_column("File", style="cyan", no_wrap=True)
    table.add_column("Code", justify="right")
    table.add_column("Docstring", justify="right")
    table.add_column("Other", justify="right")

    command_level_success = True
    for key, value in data.items():

        code, doc, other = value.get("code"), value.get("docstring"), value.get("other")
        success = code <= code_threshold
        emojy, msg, color, log_lvl = interpret(success)
        # lbl = f"[{color}]{key}[/{color}]"
        lbl = f"{key}"
        code, doc, other = str(code), str(doc), str(other)
        if not success:
            code = f"[{color}]{code}[/{color}]"
        row_values = [lbl, code,doc,other,]
        table.add_row(*row_values)

        command_level_success &= success

    console.print(table)

    return command_level_success




"""Perform assertions on HTTP response from the test client."""
import http
import importlib
#import json
import time
from typing import Any, cast, Protocol, Mapping, Union
import requests

from  scenery import config, logger
from scenery.common import (
    SceneryTestCase, 
    RemoteBackendTestCase,
    RemoteFrontendTestCase,
    Framework, DjangoBackendTestCase, DjangoFrontendTestCase
    )
from scenery.manifest import Take, Check, DirectiveCommand, DomArgument

import bs4
from selenium import webdriver

if config.framework == Framework.DJANGO.value :
    import django.http



# RESPONSE PROTOCOL
###################

class ResponseProtocol(Protocol):
    """A protocol for HTTP responses, covering both basic Django http response and from Selenium Driver."""

    @property
    def status_code(self) -> int:
        """The HTTP status code of the response."""

    @property
    def headers(self) -> Mapping[str, str]:
        """The headers of the response."""

    @property
    def content(self) -> Any:
        """The content of the response."""

    # @property
    # def json(self) -> str | None:
    #     """The json of the response."""
    #     raise NotImplementedError

    # @property
    # def charset(self) -> str | None:
    #     """The charset of the response."""

    # def has_header(self, header_name: str) -> bool:
    #     """Check if the response has a specific header."""

    # def __getitem__(self, header_name: str) -> str: ...

    # def __setitem__(self, header_name: str, value: str) -> None: ...



# NOTE mad: we do not declare any django.http.HttpResponse child 
# as this is the whole point of protocols to avoid painful inheritance

class SeleniumResponse(ResponseProtocol):
    """A response wrapper class for Selenium WebDriver operations.

    This class implements the ResponseProtocol interface for Selenium WebDriver,
    providing access to response data like headers, content, and charset. Note that
    some HTTP-specific features like status codes are not available through Selenium.

    Args:
        driver (webdriver.Chrome): The Selenium Chrome WebDriver instance to wrap.

    Attributes:
        driver (webdriver.Chrome): The wrapped Selenium WebDriver instance.
        _headers (dict[str, str]): Dictionary storing response headers.

    Properties:
        status_code (int): Not implemented for Selenium responses.
        headers (dict[str, str]): Dictionary of response headers.
        content (Any): Page source of the current webpage.
        charset (str): Character encoding of the response.

    Methods:
        has_header(header_name: str) -> bool: Check if a header exists in the response.
    """

    def __init__(
        self, 
        driver: webdriver.Chrome,
        ) -> None:
        self.driver = driver
        self._headers : dict[str, str] = {}

    @property
    def status_code(self) -> int:
        """Not implemented for Selenium responses."""
        # NOTE mad: this is probably hard to solve in general
        # we can't use Selenium for the status code
        raise NotImplementedError
    
    @property
    def headers(self) -> dict[str, str]:
        """Dictionary of response headers."""
        return self._headers
    
    @property
    def content(self) -> Any:
        """Page source of the current webpage."""
        return self.driver.page_source
    
    @property
    def charset(self) -> str | None:
        """Character encoding of the response."""
        # return None
        return None
    
    def has_header(self, header_name: str) -> bool:
        """Check if a header exists in the response."""
        return header_name in self._headers
    
    def __getitem__(self, header_name: str) -> str:
        return self._headers[header_name]
    
    def __setitem__(self, header_name: str, value: str) -> None:
        self._headers[header_name] = value


# RESPONSE CHECKER
##################

class Checker:
    """A utility class for performing HTTP requests and assertions on responses.

    This class provides static methods to execute HTTP requests and perform
    various checks on the responses, as specified in the test manifests.
    """

    selenium_instructions = None

    @classmethod
    def load_selenium_instructions(cls):
        instructions_file_path = config.selenium_instructions
        if instructions_file_path:
            spec = importlib.util.spec_from_file_location("dynamic_module", instructions_file_path)
            module = importlib.util.module_from_spec(spec)
            # sys.modules[module_name] = module  # Optional: add to sys.modules
            spec.loader.exec_module(module)
        else:
            module = None
        cls.selenium_instructions = module


    # NOTE mad: the first functions take a Take 
    # as argument to retrieve the server respone
    # The next functions take the response protocols
    # and potentially other arguments to perform checks

    # COLLECT RESPONSE
    ##################
    if config.framework == Framework.DJANGO.value:

        @staticmethod
        def get_django_client_response(
            testcase: DjangoBackendTestCase, take: Take
        ) -> django.http.HttpResponse:
            """Execute an HTTP request based on the given HttpTake object.

            Args:
                testcase (BackendDjangoTestCase): The Django testcase instance.
                take (scenery.manifest.HttpTake): The HttpTake object specifying the request details.

            Returns:
                django.http.HttpResponse: The response from the HTTP request.

            Raises:
                NotImplementedError: If the HTTP method specified in the take is not implemented.
            """
            logger.debug(Checker.get_django_client_response.__name__)
            logger.debug(f"{take.url=}")
            logger.debug(f"{take.method=}")
            if take.data:
                logger.debug(take.data)

            if take.method == http.HTTPMethod.GET:
                response = testcase.client.get(
                    take.url,
                    take.data,
                )
            elif take.method == http.HTTPMethod.POST:
                response = testcase.client.post(
                    take.url,
                    take.data,
                )
            else:
                raise NotImplementedError(take.method)

            # FIXME mad: this one is a bit puzzling to me
            # running mypy I get:
            # Incompatible return value type (got "_MonkeyPatchedWSGIResponse", expected "HttpResponse")
            return response # type: ignore[return-value]
    else:
        @staticmethod
        def get_django_client_response(testcase : DjangoBackendTestCase, take: Take):
            raise ValueError("Django is required for this method")
        
    @classmethod
    def get_selenium_response(
        cls, testcase: Union[RemoteFrontendTestCase, DjangoFrontendTestCase], take: Take
    ) -> SeleniumResponse:
        """Create a SeleniumResponse by executing a request through Selenium WebDriver.

        This function handles both GET and POST requests through Selenium. For POST requests,
        it dynamically loads and executes request handlers from a configured Selenium module.

        Args:
            testcase (FrontendDjangoTestCase): The test case instance containing
                the Selenium WebDriver and live server URL.
            take (Take): The request specification containing method, URL, and data
                for the request to be executed.

        Returns:
            SeleniumResponse: A wrapper containing the response from the Selenium-driven request.

        Raises:
            ImportError: If the SCENERY_POST_REQUESTS_INSTRUCTIONS_SELENIUM module cannot be loaded.
            AttributeError: If a POST request handler method cannot be found in the Selenium module.

        Notes:
            - For POST requests, the handler method name is derived from the URL name by
            replacing ':' with '_' and prefixing with 'post_'.
            - The Selenium module path must be specified in the SCENERY_POST_REQUESTS_INSTRUCTIONS_SELENIUM
            environment variable.
        """

        if cls.selenium_instructions is None:
            cls.load_selenium_instructions()
        # url = testcase.live_server_url + take.url
        url = testcase.base_url + take.url

        logger.debug(Checker.get_selenium_response.__name__)
        logger.debug(f"{url=}")
        logger.debug(f"{take.method=}")
        if take.data:
            logger.debug(take.data)


        response = SeleniumResponse(testcase.driver)

        testcase.driver.get(url)
        # Wait for DOMContentLoaded event
        # NOTE mad: redundant with selenium eager mode, remove if tested
        testcase.driver.execute_script("""
            return new Promise((resolve) => {
                if (document.readyState === 'interactive' || document.readyState === 'complete') {
                    resolve();
                } else {
                    document.addEventListener('DOMContentLoaded', resolve);
                }
            });
        """)
        
        # TODO mad: improve and document
        method_name = take.url_name.replace(":", "_")
        # method_name = take.url_name.replace("-", "_")
        # method_name = take.url_name.replace(".", "_")
        method_name = method_name.replace("-", "_")
        method_name = method_name.replace(".", "_")

        if take.method == http.HTTPMethod.GET:

            method_name =  f"get_{method_name}"
            if hasattr(cls.selenium_instructions, method_name):
                get_method = getattr(cls.selenium_instructions, method_name)
                get_method(testcase, url, take.data, take.query_parameters)

        if take.method == http.HTTPMethod.POST:
            method_name =  f"post_{method_name}"
            post_method = getattr(cls.selenium_instructions, method_name)
            post_method(testcase, url, take.data, take.query_parameters)

        return response 
    
    @staticmethod
    def get_http_response(testcase: RemoteBackendTestCase, take: Take) -> requests.Response:

        url = testcase.base_url + take.url

        logger.debug(Checker.get_http_response.__name__)
        logger.debug(f"{url=}")
        logger.debug(f"{take.method=}")
        if take.data:
            logger.debug(take.data)

        if take.method == http.HTTPMethod.GET:
            response = testcase.session.get(
                url,
                data=take.data,
                headers=testcase.headers,
            )
        elif take.method == http.HTTPMethod.POST:
            response = testcase.session.post(
                testcase.base_url + take.url,
                take.data,
                headers=testcase.headers,
            )
        else:
            raise NotImplementedError(take.method)

        return response

    # CHECKS
    ################################

    @staticmethod
    def exec_check(
        testcase: SceneryTestCase,
        response: ResponseProtocol,
        check: Check,
    ) -> None:
        """Execute a specific check on an HTTP response.

        This method delegates to the appropriate check method based on the instruction
        specified in the HttpCheck object.

        Args:
            testcase (DjangoTestCase): The Django test case instance.
            response (ResponseProtocol): The response to check.
            check (scenery.manifest.HttpCheck): The check to perform on the response.

        Raises:
            NotImplementedError: If the check instruction is not implemented.
        """

        logger.debug(check)

        if check.instruction == DirectiveCommand.STATUS_CODE:
            Checker.check_status_code(testcase, response, check.args)
        elif check.instruction == DirectiveCommand.REDIRECT_URL:
            Checker.check_redirect_url(testcase, response, check.args)
        elif check.instruction == DirectiveCommand.COUNT_INSTANCES:
            Checker.check_count_instances(testcase, response, check.args)
        elif check.instruction == DirectiveCommand.DOM_ELEMENT:
            Checker.check_dom(testcase, response, check.args)
        elif check.instruction == DirectiveCommand.JSON:
            Checker.check_json(testcase, response, check.args)
        elif check.instruction == DirectiveCommand.FIELD_OF_INSTANCE:
            Checker.check_field_of_instance(testcase, response, check.args)
        # NOTE mad: do not erase
        # elif check.instruction == scenery.manifest.DirectiveCommand.JS_VARIABLE:
        #     Checker.check_js_variable(testcase, response, check.args)
        # elif check.instruction == scenery.manifest.DirectiveCommand.JS_STRINGIFY:
        #     Checker.check_js_stringify(testcase, response, check.args)
        else:
            raise NotImplementedError(check)
        

    @staticmethod
    def check_status_code(
        testcase: SceneryTestCase,
        response: ResponseProtocol,
        args: int,
    ) -> None:
        """Check if the response status code matches the expected code.

        Args:
            testcase (DjangoTestCase): The Django test case instance.
            response (ResponseProtocol): The HTTP response to check.
            args (int): The expected status code.
        """
        testcase.assertEqual(
            response.status_code,
            args,
            f"Expected status code {args}, but got {response.status_code}",
        )

    @staticmethod
    def check_redirect_url(
        testcase: SceneryTestCase,
        response: ResponseProtocol,
        args: str,
    ) -> None:
        """Check if the response redirect URL matches the expected URL.

        Args:
            testcase (DjangoTestCase): The Django test case instance.
            response (ResponseProtocol): The HTTP response to check.
            args (str): The expected redirect URL.
        """
        # NOTE mad: this will fail when we try with frontend for login etc... 
        # but I skip those kind of test in the method builder
        if config.framework == Framework.DJANGO.value and isinstance(response, django.http.HttpResponseRedirect):
            testcase.assertEqual(
                response.url,
                args,
                f"Expected redirect URL '{args}', but got '{response.url}'",
            )
        elif isinstance(response, requests.Response):
            testcase.assertEqual(
                response.url,
                args,
                f"Expected redirect URL '{args}', but got '{response.url}'",
            )
        else:
            testcase.fail(f"Expected response of type HttpResponseRedirect or requests.Response but got {type(response)}")


    @staticmethod
    def check_count_instances(
        testcase: SceneryTestCase,
        response: ResponseProtocol,
        args: dict,
    ) -> None:
        """Check if the count of model instances matches the expected count.

        Args:
            testcase (DjangoTestCase): The Django test case instance.
            response (ResponseProtocol): The HTTP response (not used in this check).
            args (dict): A dictionary containing 'model' (the model class) and 'n' (expected count).
        """
        instances = list(args["model"].objects.all())
        testcase.assertEqual(
            len(instances),
            args["n"],
            f"Expected {args['n']} instances of {args['model'].__name__}, but found {len(instances)}",
        )

    @staticmethod
    def check_dom(
        testcase: SceneryTestCase,
        response: ResponseProtocol,
        args: dict[DomArgument, Any],
    ) -> None:
        """Check for the presence and properties of DOM elements in the response content.

        This method uses BeautifulSoup to parse the response content and perform various
        checks on DOM elements as specified in the args dictionary.

        Args:
            testcase (DjangoTestCase): The Django test case instance.
            response (django.ResponseProtocol): The HTTP response to check.
            args (dict): A dictionary of DomArgument keys and their corresponding values,
                         specifying the checks to perform.

        Raises:
            ValueError: If neither 'find' nor 'find_all' arguments are provided in args.
        """

        # TODO: this function is getting out of control


        # NOTE mad: this is incredibly important for the frontend test
        time.sleep(1)

        soup = bs4.BeautifulSoup(response.content, "html.parser")

        # Apply the scope
        if scope := args.get(DomArgument.SCOPE):
            scope_result = soup.find(**scope)
            testcase.assertIsNotNone(
                scope,
                f"Expected to find an element matching {args[DomArgument.SCOPE]}, but found none",
            )
        else:
            scope_result = soup

        # FIXME mad: we inforce type checking by regarding bs4 objects as Tag
        scope_result = cast(bs4.Tag, scope_result)

        # Locate the element(s)
        if args.get(DomArgument.FIND_ALL):
            dom_elements = scope_result.find_all(**args[DomArgument.FIND_ALL])
            testcase.assertGreaterEqual(
                len(dom_elements),
                1,
                f"Expected to find at least one element matching {args[DomArgument.FIND_ALL]}, but found none",
            )
        elif args.get(DomArgument.FIND):
            dom_element = scope_result.find(**args[DomArgument.FIND])
            testcase.assertIsNotNone(
                dom_element,
                f"Expected to find an element matching {args[DomArgument.FIND]}, but found none",
            )
            dom_elements = bs4.ResultSet(source=bs4.SoupStrainer(), result=[dom_element])
        else:
            raise ValueError("Neither find of find_all argument provided")
        # FIXME mad: as I enforce the results to be a bs4.ResultSet[bs4.Tag] above
        dom_elements = cast(bs4.ResultSet[bs4.Tag], dom_elements)

        # Perform the additional checks
        if count := args.get(DomArgument.COUNT):
            testcase.assertEqual(
                len(dom_elements),
                count,
                f"Expected to find {count} elements, but found {len(dom_elements)}",
            )
        for dom_element in dom_elements:
            if text := args.get(DomArgument.TEXT):
                testcase.assertEqual(
                    dom_element.text,
                    text,
                    f"Expected element text to be '{text}', but got '{dom_element.text}'",
                )
            if attribute := args.get(DomArgument.ATTRIBUTE):

                if value := attribute.get("value"):
                    # TODO mad: should this move to manifest parser?
                    # in manifest _format_dom_element should be used here, or even before and just disappear
                    if isinstance(value, (str, list)):
                        pass
                    elif isinstance(value, int):
                        value = str(value)
                    else:
                        raise TypeError(
                            f"attribute value can only by `str` or `list[str]` not '{type(value)}'"
                        )
                    testcase.assertEqual(
                        dom_element[attribute["name"]],
                        value,
                        f"Expected attribute '{attribute['name']}' to have value '{value}', but got '{dom_element[attribute['name']]}'",
                    )
                elif regex := attribute.get("regex"):

                    testcase.assertRegex(
                        dom_element[attribute["name"]],
                        regex,
                        f"Expected attribute '{attribute['name']}' to match regex '{regex}', but got '{dom_element[attribute['name']]}'",
                    )

                
                # NOTE: do note erase
                # if exepected_value_from_ff := attribute.get("json_stringify"):
                    
                    

                    # if isinstance(testcase, DjangoFrontendTestCase):
                    #     # NOTE mad: we cannot anotate it.
                    #     value_from_ff = testcase.driver.execute_script( # type: ignore[no-untyped-call]
                    #     f"return JSON.stringify({dom_element[attribute['name']]})"
                    # )
                    # else:
                    #     raise Exception("json_stringify can only be called for frontend tests")

                    # if exepected_value_from_ff == "_":
                    #     # NOTE mad: this means we only want to check the value is a valid json string
                    #     pass
                    # else:
                    #     value_from_ff = json.loads(value_from_ff)
                    #     testcase.assertEqual(
                    #         value_from_ff,
                    #         exepected_value_from_ff,
                    #         f"Expected attribute '{attribute['name']}' to have value '{exepected_value_from_ff}', but got '{value_from_ff}'",
                    #     )
                
    @staticmethod
    def check_json(
        testcase: SceneryTestCase,
        response: ResponseProtocol,
        args: dict,
    ) -> None:

        if config.framework == Framework.DJANGO.value :
            # print(f"{response.json()=}")
            testcase.assertIsInstance(response, django.http.JsonResponse)
            # FIXME
            response_json = response.json() # type: ignore [attr-defined]
            testcase.assertEqual(response_json[args["key"]], args["value"])

        else:
            response_json = response.json()
            testcase.assertEqual(response_json[args["key"]], args["value"])
            # print(args)
            # print(response_json)
            # testcase.assertEqual(dir(response), "lala")



    @staticmethod
    def check_field_of_instance(
        testcase: SceneryTestCase,
        response: ResponseProtocol,
        args: dict,
    ) -> None:
        

        instances = list(args["find"]["model"].objects.all())
        if len(instances) != 1:
            testcase.fail(f"Checking the {args['field']} field of {args['find']['model']} requires that there is a single instance in the db, but found {len(instances)}.")
        
        instance = instances[0]
        field_value = getattr(instance, args["field"])
        testcase.assertEqual(
            field_value, 
            args["value"],
            f"{args['find']['model'].__name__}.{args['field']} = {field_value} but expected {args['value']}"
        )

        # testcase.assertEqual(
        #     len(instances),
        #     args["n"],
        #     f"Expected {args['n']} instances of {args['model'].__name__}, but found {len(instances)}",
        # )

# NOTE mad: do not erase
    # def check_js_variable(self, testcase: DjangoFrontendTestCase, args: dict) -> None:
    #     """
    #     Check if a JavaScript variable has the expected value.
    #     Args:
    #         testcase (DjangoTestCase): The Django test case instance.
    #         args (dict): The arguments for the check.
    #     """

    #     variable_name = args["name"]
    #     expected_value = args["value"]
    #     actual_value = testcase.driver.execute_script(
    #         f"return {variable_name};"
    #     )
    #     testcase.assertEqual(
    #         actual_value,
    #         expected_value,
    #         f"Expected JavaScript variable '{variable_name}' to have value '{expected_value}', but got '{actual_value}'",
    #     )


"""General functions and classes used by other modules."""

import argparse
import collections
import enum
import os
import unittest
import requests
from typing import Any, Iterable, List, TypeVar, Union

from scenery import config

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
import yaml



###################
# TYPES
###################

# NOTE mad: this is here to prevent circular import and still use those types
# either to check isinstance(x, cls) (see reponse_checker for instance),
# and type checking



class RemoteBackendTestCase(unittest.TestCase):
    """A TestCase for backend testing on a remote server."""
    mode: str
    session: requests.Session
    base_url: str
    headers: dict[str, str]

class RemoteFrontendTestCase(unittest.TestCase):
    """A TestCase for backend testing on a remote server."""
    mode: str
    driver: webdriver.Chrome
    # session: requests.Session
    base_url: str
    headers: dict[str, str]

class LoadTestCase(unittest.TestCase):
    """A TestCase for load testing on a remote server."""
    mode: str
    session: requests.Session
    headers: dict[str, str]
    base_url: str
    data: dict[str, List[dict[str, int|float]]]
    users:int
    requests_per_user:int


# NOTE mad: proper implementation is in django utils
class DjangoBackendTestCase:
    def __init__(self, *args, **kwargs):
            ImportError("django required")

class DjangoFrontendTestCase:
    def __init__(self, *args, **kwargs):
            ImportError("django required")


class Framework(enum.Enum):
    DJANGO = "django"


SceneryTestCaseTypes = Union[RemoteBackendTestCase, RemoteFrontendTestCase, LoadTestCase]

SceneryTestCase = TypeVar("SceneryTestCase", bound=SceneryTestCaseTypes)



###################
# SELENIUM
###################


def get_selenium_driver(headless: bool, page_load_timeout: int = 30) -> webdriver.Chrome:
    """Return a Selenium WebDriver instance configured for Chrome.

    Args:
        headless (bool): Whether to run Chrome in headless mode.
        page_load_timeout (int): Maximum time in seconds to wait for the page to load.
            If the timeout is exceeded, a TimeoutException will be thrown.

    Returns:
        webdriver.Chrome: Configured Chrome WebDriver instance.
    """
    chrome_options = Options()
    chrome_options.add_argument('--no-sandbox')
    chrome_options.add_argument('--disable-dev-shm-usage')
    chrome_options.add_argument('--disable-gpu')
    chrome_options.add_argument('--window-size=1920,1080')
    chrome_options.add_argument('--user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36')
    # NOTE mad: used to wait until domcontent loaded
    # see: https://www.selenium.dev/documentation/webdriver/drivers/options/
    chrome_options.page_load_strategy = 'eager' 
    if headless:
        chrome_options.add_argument("--headless=new")         # NOTE mad: For newer Chrome versions
        # chrome_options.add_argument("--headless")           
    driver = webdriver.Chrome(options=chrome_options)  #  service=service
    driver.implicitly_wait(10)
    driver.set_page_load_timeout(page_load_timeout)  # Set timeout for page load
    return driver



########
# YAML #
########


def read_yaml(filename: str) -> Any:
    """Read and parse a YAML file.

    Args:
        filename (str): The path to the YAML file to be read.

    Returns:
        Any: The parsed content of the YAML file.

    Raises:
        yaml.YAMLError: If there's an error parsing the YAML file.
        IOError: If there's an error reading the file.
    """
    with open(filename, "r") as f:
        return yaml.safe_load(f)


def iter_on_manifests(args: argparse.Namespace) -> Iterable[str]:
    for filename in os.listdir(config.manifests_folder):
        if args.manifest is not None and filename.replace(".yml", "") != args.manifest:
            continue

        yield filename



##################
# UNITTEST
##################


def serialize_unittest_result(result: unittest.TestResult) -> collections.Counter:
    """Serialize a unittest.TestResult object into a dictionary.

    Args:
        result (unittest.TestResult): The TestResult object to serialize.

    Returns:
        dict: A dictionary containing the serialized TestResult data.
    """
    d = {
        attr: getattr(result, attr)
        for attr in [
            "failures",
            "errors",
            "testsRun",
            "skipped",
            "expectedFailures",
            "unexpectedSuccesses",
        ]
    }
    d = {key: len(val) if isinstance(val, list) else val for key, val in d.items()}

    d["failed assertions"] = d["failures"]
    d["errors during execution"] = d["errors"]
    d["failed manifests"] = int(d["failures"] > 0)
    d["successful manifests"] = int(d["testsRun"])
    # d[""] = int(d["errors"] > 0)
    d.pop("errors")
    d.pop("failures")
    d.pop("testsRun")
    return collections.Counter(d)




